package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.signature.InvalidSignatureException
import com.ionspin.kotlin.crypto.signature.Signature
import kotlinx.datetime.Instant
import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import net.sergeych.crypto2.Signing.Companion.pair
import net.sergeych.utools.now

interface VerifyingKey {
    val id: KeyId

    /**
     * Verify the signature and return true if it is correct.
     */
    fun verify(signature: UByteArray, message: UByteArray): Boolean
}

interface SigningKey {
    val publicKey: Signing.PublicKey
    fun sign(message: UByteArray): UByteArray
    fun seal(message: UByteArray, expiresAt: Instant? = null): Seal
}

/**
 * Keys in general: public, secret and later symmetric too.
 * Keys could be compared to each other for equality and used
 * as a Map keys (not sure about js).
 *
 * Use [pair] to create new keys.
 */
@Serializable
sealed class Signing {
    abstract val id: KeyId

    abstract val packed: UByteArray

    override fun equals(other: Any?): Boolean {
        return other is Signing && other.packed contentEquals packed
    }

    override fun hashCode(): Int {
        return packed.contentHashCode()
    }

    override fun toString(): String = packed.encodeToBase64Url()

    /**
     * Public key to verify signatures only
     */
    @Serializable
    @SerialName("p")
    class PublicKey(override val packed: UByteArray) : Signing(), VerifyingKey {
        /**
         * Verify the signature and return true if it is correct.
         */
        override fun verify(signature: UByteArray, message: UByteArray): Boolean = try {
            Signature.verifyDetached(signature, message, packed)
            true
        } catch (_: InvalidSignatureException) {
            false
        }

        override fun toString(): String = "Pub:${super.toString()}"

        override val id: KeyId by lazy {
            KeyId(KeysMagickNumbers.defaultSigning, packed)
        }
    }

    /**
     * Secret key to sign only
     */
    @Serializable
    @SerialName("s")
    class SecretKey(override val packed: UByteArray) : Signing(), SigningKey {

        override val publicKey: PublicKey by lazy {
            PublicKey(Signature.ed25519SkToPk(packed))
        }

        override fun sign(message: UByteArray): UByteArray = Signature.detached(message, packed)

        override fun seal(message: UByteArray, expiresAt: Instant?): Seal =
            Seal.create(this, message, now(), expiresAt)

        override fun toString(): String = "Sct:${super.toString()}"

        override val id: KeyId
            get() = publicKey.id
    }

    companion object {
        data class Pair(val secretKey: SecretKey, val publicKey: PublicKey)

        fun pair(): Pair {
            val p = Signature.keypair()
            return Pair(SecretKey(p.secretKey), PublicKey(p.publicKey))
        }
    }
}

open class IllegalSignatureException(text: String="signed data is tampered or signature is corrupted")
    : IllegalStateException(text)

class ExpiredSignatureException(text: String): IllegalSignatureException(text)