package net.sergeych.crypto2

import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient

@Serializable
sealed class UniversalKey {
    abstract val id: KeyId
    @Transient
    open val canEncrypt = false
    @Transient
    open val canDecrypt = false
    @Transient
    open val canSign = false
    @Transient
    open val canVerify = false

    @Serializable
    @SerialName("sym")
    data class Symmetric(val key: SymmetricKey) : UniversalKey(), EncryptingKey by key, DecryptingKey by key {
        @Transient
        override val id: KeyId = key.id

        @Transient
        override val nonceBytesLength: Int = key.nonceBytesLength

        override fun toString() = "U.Sym:$id"

        @Transient
        override val canDecrypt: Boolean = true
        @Transient
        override val canEncrypt: Boolean = true
    }

    @Serializable
    @SerialName("ssn")
    data class Session(val key: SafeKeyExchange.SessionKey) : UniversalKey(), EncryptingKey by key,
        DecryptingKey by key {
        @Transient
        override val id: KeyId = key.id
        @Transient
        override val nonceBytesLength: Int = key.nonceBytesLength
        @Transient
        override val canDecrypt: Boolean = true
        @Transient
        override val canEncrypt: Boolean = true
    }

    @Serializable
    @SerialName("sec")
    data class Secret(val key: Asymmetric.SecretKey) : UniversalKey(), DecryptingKey by key {
        override val id: KeyId by lazy { key.id }
        override fun toString() = "U.Sec:$id"
        @Transient
        override val canDecrypt: Boolean = true
    }

    @Serializable
    @SerialName("sig")
    data class Signing(val key: net.sergeych.crypto2.Signing.SecretKey) : UniversalKey() {
        override val id: KeyId by lazy { key.id }
        override fun toString() = "U.Sig:$id"
        @Transient
        override val canDecrypt: Boolean = true
    }

    @Serializable
    @SerialName("ver")
    data class Verifying(val key: net.sergeych.crypto2.Signing.PublicKey) : UniversalKey() {
        override val id: KeyId by lazy { key.id }
        override fun toString() = "U.Sig:$id"
        @Transient
        override val canDecrypt: Boolean = true
    }



    companion object {
        fun from(key: DecryptingKey): UniversalKey =
            when (key) {
                is UniversalKey -> key
                is Asymmetric.SecretKey -> Secret(key)
                is SymmetricKey -> Symmetric(key)
                is SafeKeyExchange.SessionKey -> Session(key)
                else -> throw UnsupportedOperationException("can't create universal key from ${key::class.simpleName}")
            }
    }

}
